/***************************************************************************
 *   SDL_rwfd.c -- SDL_RWops stream implimentation for UNIX file           *
 *     descriptors                                                         *
 *                                                                         *
 *   Copyright (C) 2004 by Tyler Montbriand                                *
 *   tsm@accesscomm.ca                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <sys/types.h>

#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <SDL/SDL_error.h>
#include "SDL_rwfd.h"

#define SDL_FD_TYPE 0xdead0001
#define STDIN_FD    0
#define STDOUT_FD   1
#define STDERR_FD   2

static int sdl_fd_seek (SDL_RWops *c, int offset, int whence);
static int sdl_fd_read (SDL_RWops *c, void *vp, int size, int maxnum);
static int sdl_fd_write(SDL_RWops *c, const void *vp, int size, int maxnum);
static int sdl_fd_close(SDL_RWops *c);
static int sdl_fd_close_static(SDL_RWops *c);

static SDL_RWops rw_fd_stdin={ sdl_fd_seek
                             , sdl_fd_read
                             , sdl_fd_write
                             , sdl_fd_close_static
                             , SDL_FD_TYPE
                             , {0,(FILE *)STDIN_FD} };

static SDL_RWops rw_fd_stdout={sdl_fd_seek
                             , sdl_fd_read
                             , sdl_fd_write
                             , sdl_fd_close_static
                             , SDL_FD_TYPE
                             , {0,(FILE *)STDOUT_FD} };

static SDL_RWops rw_fd_stderr={sdl_fd_seek
                             , sdl_fd_read
                             , sdl_fd_write
                             , sdl_fd_close_static
                             , SDL_FD_TYPE
                             , {0,(FILE *)STDERR_FD} };

SDL_RWops *rwin =&rw_fd_stdin, 
          *rwout=&rw_fd_stdout,
          *rwerr=&rw_fd_stderr;

SDL_RWops *SDL_RWFromFD(int fd, int autoclose)
{
  SDL_RWops *rw;

  if(fd<0)
  {
    SDL_SetError("SDL_RWFromFD:  Invalid file # %d",fd);
    return(NULL);
  }

  rw=SDL_AllocRW();
  if(rw==NULL)
  {
    SDL_SetError("Out of memory");
    return(NULL);
  }
  memset(rw,0,sizeof(SDL_RWops));
  rw->seek=sdl_fd_seek;
  rw->read=sdl_fd_read;
  rw->write=sdl_fd_write;
  rw->close=sdl_fd_close;
  rw->type=SDL_FD_TYPE;
  rw->hidden.stdio.autoclose=autoclose;
  rw->hidden.stdio.fp=(FILE *)(long int)fd;

  return(rw);
}

static int sdl_fd_seek (SDL_RWops *c, int offset, int whence)
{
  return(lseek((int)(long int)(c->hidden.stdio.fp),offset,whence));
}

static int sdl_fd_read (SDL_RWops *c, void *vp, int size, int maxnum)
{
  int bytes=read((int)(long int)(c->hidden.stdio.fp),vp,size*maxnum);
  if(bytes%size) // Didn't manage to read all the records, realign
  {
    int off=-(bytes%size);
    lseek((int)(long int)(c->hidden.stdio.fp),-(bytes%size),SEEK_CUR);
  }

  if(bytes>=0) return(bytes/size);
  else         return(-1);
}

static int sdl_fd_write(SDL_RWops *c, const void *vp, int size, int maxnum)
{
  int bytes=write((int)(long int)(c->hidden.stdio.fp),vp,size*maxnum);
  if(bytes%size) // Didn't manage to read all the records, realign
  {
    int off=-(bytes%size);
    lseek((int)(long int)(c->hidden.stdio.fp),-(bytes%size),SEEK_CUR);
  }

  if(bytes>=0) return(bytes/size);
  else         return(-1);
}

static int sdl_fd_close(SDL_RWops *c)
{
  int retval=0;
  if(c==NULL)
  {
    SDL_SetError("sdl_fd_close: NULL context");
    return(-1);
  }
  else if(c->type!=SDL_FD_TYPE)
  {
    SDL_SetError("sdl_fd_close: context is not SDL_FD_TYPE");
    return(-1);
  }

  if(c->hidden.stdio.autoclose)
    retval=close((int)(long int)(c->hidden.stdio.fp));

  SDL_FreeRW(c);
  return(retval);  
}

static int sdl_fd_close_static(SDL_RWops *c)
{
  SDL_SetError("close is undefined for static sdl_fd constructs");
  return(-1);
}

#ifdef RW_FD_TEST
#include "SDL_rwprintf.h"
int main(int argc, char *argv[])
{
  char buf[80]={'\0'};
  int bytes;
  rwprintf(rwerr,"Hello World %d\n",42);
  bytes=SDL_RWread(rwin,buf,1,79);
  if(bytes>0) buf[bytes]='\0';
  rwprintf(rwout,"Read %d bytes: %s\n",bytes,buf);
  
  return(0);
}
#endif/*RW_FD_TEST*/
