/***************************************************************************
 *   Temporary file and memory buffer SDL_RWops stream implimentation      *
 *                                                                         *
 *   Copyright (C) 2004 by Tyler Montbriand                                *
 *   tsm@accesscomm.ca                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <SDL/SDL_error.h>
#include "SDL_rwtmp.h"

#define SDL_TMP_TYPE 0xdead0006
#define SDL_TMP_BLOCK 4096 // Memory block size

static int sdl_tmp_seek (SDL_RWops *c, int offset, int whence);
static int sdl_tmp_read (SDL_RWops *c, void *vp, int size, int maxnum);
static int sdl_tmp_write(SDL_RWops *c, const void *vp, int size, int maxnum);
static int sdl_tmp_close(SDL_RWops *c);

SDL_RWops *SDL_RWTmp()
{
  SDL_RWops *rw;

  rw=SDL_AllocRW();
  if(rw==NULL)
  {
    SDL_SetError("Out of memory");
    return(NULL);
  }

  memset(rw,0,sizeof(SDL_RWops));
  rw->seek=sdl_tmp_seek;
  rw->read=sdl_tmp_read;
  rw->write=sdl_tmp_write;
  rw->close=sdl_tmp_close;
  rw->type=SDL_TMP_TYPE;
  rw->hidden.mem.base=NULL;
  rw->hidden.mem.here=NULL;
  rw->hidden.mem.stop=NULL;
  return(rw);
}

static int sdl_tmp_seek (SDL_RWops *c, int offset, int whence)
{
  Uint8 *off;

  switch(whence)
  {
  case SEEK_SET: /* Seek from beginning of buffer */
    off=(c->hidden.mem.base)+offset;
    break;
  case SEEK_CUR: /* Seek from current position */
    off=(c->hidden.mem.here)+offset;
    break;
  case SEEK_END: /* Seek from end of file */
    off=(c->hidden.mem.stop)+offset;
    break;
  default:
    SDL_SetError("sdl_tmp_seek:  whence should be SEEK_SET, SEEK_CUR, SEEK_SET");
    return(-1);
  }

  if(off>(c->hidden.mem.stop))      off=c->hidden.mem.stop;
  else if(off<(c->hidden.mem.base)) off=c->hidden.mem.base;

  c->hidden.mem.here=off;

  return((int)(long int)(off-(c->hidden.mem.base)));
}

static int sdl_tmp_read (SDL_RWops *c, void *vp, int size, int maxnum)
{
  long int left=(long int)((c->hidden.mem.stop)-(c->hidden.mem.here));
  long int blocks=(left/size);

  if(blocks==0)     return(0);

  if(blocks>maxnum) blocks=maxnum;

  memcpy(vp,c->hidden.mem.here,blocks*size);
  c->hidden.mem.here+=(blocks*size);
  return(blocks);
}

static int sdl_tmp_resize(SDL_RWops *c, long int more)
{
  // get original parameters
  Uint8 *base=c->hidden.mem.base,*nbase;
  long int pos=(long int)(c->hidden.mem.here-c->hidden.mem.base);
  long int len=(long int)(c->hidden.mem.stop-c->hidden.mem.base);
  long int nlen=len+more;

  if(nlen%SDL_TMP_BLOCK)
  {
    nlen-=nlen%SDL_TMP_BLOCK;
    nlen+=SDL_TMP_BLOCK;
  }

  nbase=realloc(base,nlen);

  if(nbase==NULL)
  {
    SDL_SetError("sdl_tmp_resize:  Out of memory");
    return(-1);
  }

  c->hidden.mem.base=nbase;
  c->hidden.mem.here=nbase+pos;
  c->hidden.mem.stop=nbase+nlen;

  return(0);
}

static int sdl_tmp_write(SDL_RWops *c, const void *vp, int size, int maxnum)
{
  long int left=(long int)((c->hidden.mem.stop)-(c->hidden.mem.here));

  if(left<(size*maxnum)) // Attempt to expand memory before copying
  {
    int more=(size*maxnum)-left;

    if(sdl_tmp_resize(c,more)<0) // Cannot resize, but can still write
      maxnum=(left/size);
  }

  // Copy all in one go!
  if(maxnum) memcpy(c->hidden.mem.here,vp,maxnum*size);
  return(maxnum);
}

static int sdl_tmp_close(SDL_RWops *c)
{
  int retval=0;
  if(c==NULL)
  {
    SDL_SetError("sdl_tmp_close: NULL context");
    return(-1);
  }
  else if(c->type!=SDL_TMP_TYPE)
  {
    SDL_SetError("sdl_tmp_close: context is not SDL_TMP_TYPE");
    return(-1);
  }

  if(c->hidden.mem.base!=NULL) free(c->hidden.mem.base);

  SDL_FreeRW(c);
  return(retval);  
}

int SDL_RWTmp_glob(SDL_RWops *c,SDL_RWops *src, int size, int maxnum)
{
  int blocks=0;

  if(c->type!=SDL_TMP_TYPE)
  {
    SDL_SetError("SDL_RWTmp_glob:  first parameter must be SDL_RWTmp buffer");
    return(-1);
  }

  while(blocks<maxnum)
  {
//    void *cur;
    long int left=(long int)((c->hidden.mem.stop)-(c->hidden.mem.here));

    if(left<size)
    {
      if(sdl_tmp_resize(c,left-size)<0)
        return(blocks);
    }

    if(SDL_RWread(src,c->hidden.mem.here,size,1)!=1)
      return(blocks);
    c->hidden.mem.here+=size;
    blocks++;
  }

  return(blocks);
}

SDL_RWops *SDL_RWTmpFile()
{
  SDL_RWops *rw;
  FILE *fp=tmpfile();
  if(fp==NULL) return(NULL);
  rw=SDL_RWFromFP(fp,1);
  if(rw==NULL)
  {
    fclose(fp);
    return(NULL);
  }

  return(rw);
}
